% Implementation of an interior point method 
clear variables
close all
clc

import casadi.*
% Problem definition
nv = 2;
x = MX.sym('x',nv);

% Equality contstraints
ne = 1;
g = sin(x(1)) - x(2)^2;
G = Function('G',{x},{g});
x_test = [2.,3.];
[g_test] = G({x_test});
disp(full(g_test{1}))
g_test = -8.0907;

% Inequality contstraints
ni = 1;
h = x(1)^2 + x(2)^2 - 4;
H = Function('H',{x},{h});
x_test = [2.,3.];
[h_test] = H({x_test});
disp(full(h_test{1}))

% Objective
f = (x(1)- 4)^2 + (x(2) - 4)^2;
F = Function('F',{x},{f});
x_test = [2.,3.];
[f_test] = F({x_test});
disp(full(f_test{1}))

% Create CasADi object that evaluates the Jacobian of g
Jg = Function('Jg', {x},{jacobian(g,x)});
[Jg_test] = Jg({x_test});
disp(full(Jg_test{1}))

% Create CasADi object that evaluates the Jacobian of h
Jh = Function('Jh', {x},{jacobian(h,x)});
[Jh_test] = Jh({x_test});
disp(full(Jh_test{1}))

% Create CasADi object that evaluates the Jacobian of f
Jf = Function('Jf', {x},{jacobian(f,x)});
[Jf_test] = Jf({x_test});
disp(full(Jf_test{1}))

% Create CasADi object that evaluates the Hessian of the equalities
Hg = Function('Hg', {x},{hessian(g,x)});
[Hg_test] = Hg({x_test});
disp(full(Hg_test{1}))

% Create CasADi object that evaluates the Hessian of the inequalities
Hh = Function('Hh', {x},{hessian(h,x)});
[Hh_test] = Hh({x_test});
disp(full(Hh_test{1}))

% Create CasADi object that evaluates the Hessian of the Lagrangian
Hf = Function('Hf', {x},{hessian(f,x)});
[Hf_test] = Hf({x_test});
disp(full(Hf_test{1}))

% Interior point solver
max_it = 100;
xk = [-2;4];
%xk = [1.73;0.89];
lk = 10*ones(ne,1);
vk = 10*ones(ni,1);
sk = 10*ones(ni,1);
iter = zeros(nv + ne + ni + ni,max_it);
iter(:,1) = [xk;lk;vk;sk];
tau = 2;
k_b = 1/3;
th_1 = 1.0e-8;
th_2 = 1.0e-8;
for i = 2:max_it
    % Build KKT system
    Hf_e    = Hf({xk});
    Hg_e    = Hg({xk});
    Hh_e    = Hh({xk});
    Hl      = Hf_e{1} + Hg_e{1}*lk + Hh_e{1}*vk;
    
    Jg_e    = Jg({xk});
    Jh_e    = Jh({xk});
    Jf_e    = Jf({xk});
    
    g_e     = G({xk});
    h_e     = H({xk});
    
    KKT = full([    Hl              Jg_e{1}'        Jh_e{1}'        zeros(nv,ni);   ...
                    Jg_e{1}         zeros(ne,ne)    zeros(ne,ni)    zeros(ne,ni);   ...
                    Jh_e{1}         zeros(ni,ne)    zeros(ni,ni)    eye(ni)     ;   ...
                    zeros(ni,nv)    zeros(ni,ne)    diag(sk)        diag(vk)    ]);
    
    lhs = - full([  Jf_e{1}' + Jg_e{1}'*lk + Jh_e{1}'*vk; ...
                    g_e{1}                              ; ...
                    h_e{1}  + sk                             ; ...
                    vk.*sk - tau                    ]);
    
    % Termination condition
    if norm(lhs) < th_1
        if tau < th_2
            display('Solution found!')
            break;
        else
            tau = tau*k_b;
        end
    end
        
    sol = KKT\lhs;
    % line-serach
    max_ls = 100;
    x_step  = sol(1:nv);
    l_step  = sol(nv+1:nv+ne);
    v_step  = sol(nv+ne+1:nv+ne+ni);
    s_step  = sol(nv+ne+ni+1:end);
    alpha = 1;
    k_ls = 0.9;
    min_step = 1.0e-8;
    for j=1:max_ls
        v_t = vk + alpha*v_step;
        s_t = sk + alpha*s_step;
        if (isempty(v_t(v_t <= 0)) && isempty(s_t(s_t <= 0)))
            break;
        end
        alpha = alpha*k_ls;
        if norm(alpha*[ v_step;s_step]) < min_step
            error('Line search failed! Could not find dual feasible step.')
        end
    end
    
    xk  = xk + alpha*x_step;
    lk  = lk + alpha*l_step;
    vk  = vk + alpha*v_step;
    sk  = sk + alpha*s_step;
    
    % Print some results
    display(['Iteration: ',num2str(i)]);
    display(['tau = ',num2str(tau)]);
    display(['norm(rhs) = ',num2str(norm(lhs))]);
    display(['step size = ',num2str(alpha)])
    iter(:,i) = [xk;lk;vk;sk];
end
iter = iter(:,1:i-1);
figure()
subplot(2,1,1)
plot(iter(1:2,:)')
grid on
xlabel('iterations')
ylabel('primal solution')
legend('x_1','x_2')
subplot(2,1,2)
plot(iter(3:4,:)')
grid on
xlabel('iterations')
ylabel('dual solution')
legend('\mu','\lambda')

% Plot feasible set, and iterations
figure()
pl = ezplot('sin(x) - y^2');
set(pl,'Color','red');
hold all
pl= ezplot('x^2 + y^2 - 4');
set(pl,'Color','blue');
ezcontour('(x- 4)^2 + (y - 4)^2')
plot(iter(1,:),iter(2,:),'--','Color','black')
plot(iter(1,:),iter(2,:),'o','Color','black')
title('Iterations in the primal space')
grid on


