# Implementation of an interior point method 

import casadi as C
import numpy as np

# Problem definition
nv = 2
x = C.MX.sym('x',nv)

x_test = np.array([2,3])

# Equality contstraints
ne = 1
g = np.sin(x[0]) - x[1]**2
G = C.Function('G',[x],[g])
print G([x_test]) # -8.0907;

# Inequality contstraints
ni = 1
h = x[0]**2 + x[1]**2 - 4
H = C.Function('H',[x],[h])
print H([x_test])

# Objective
f = (x[0]- 4)**2 + (x[1] - 4)**2
F = C.Function('F',[x],[f])
print F([x_test])

# Create CasADi object that evaluates the Jacobian of g
Jg = C.Function('Jg', [x],[C.jacobian(g,x)])
print Jg([x_test])

# Create CasADi object that evaluates the Jacobian of h
Jh = C.Function('Jh', [x],[C.jacobian(h,x)])
print Jh([x_test])

# Create CasADi object that evaluates the Jacobian of f
Jf = C.Function('Jf', [x],[C.jacobian(f,x)])
print Jf([x_test])

# Create CasADi object that evaluates the Hessian of the equalities
Hg = C.Function('Hg', [x],[C.hessian(g,x)[0]])
print Hg([x_test])

# Create CasADi object that evaluates the Hessian of the inequalities
Hh = C.Function('Hh', [x],[C.hessian(h,x)[0]])
print Hh([x_test])

# Create CasADi object that evaluates the Hessian of the Lagrangian
Hf = C.Function('Hf', [x],[C.hessian(f,x)[0]])
print Hf([x_test])

# Interior point solver
max_it = 100
xk = np.array([[-2,-4]]).T
lk = 10*np.ones((ne,1))
vk = 10*np.ones((ni,1))
sk = 10*np.ones((ni,1))
iters = [C.vertcat([xk,lk,vk,sk])]

tau = 0
k_b = 1.0/3
th_1 = 1.0e-3
th_2 = 1.0e-16
for i in range(1,max_it):
    # Build KKT system
    [Hf_e]    = Hf([xk])
    [Hg_e]    = Hg([xk])
    [Hh_e]    = Hh([xk])
    Hl      = Hf_e + C.mtimes(Hg_e,lk) + C.mtimes(Hh_e,vk)
    
    [Jg_e]    = Jg([xk])
    [Jh_e]    = Jh([xk])
    [Jf_e]    = Jf([xk])
    
    [g_e]     = G([xk])
    [h_e]     = H([xk])
    
    KKT = C.blockcat([[    Hl,              Jg_e.T,        Jh_e.T,        np.zeros((nv,ni))],
                    [Jg_e,        np.zeros((ne,ne)),    np.zeros((ne,ni)),    np.zeros((ne,ni))],
                    [Jh_e,         np.zeros((ni,ne)),    np.zeros((ni,ni)),    np.eye(ni)],
                    [np.zeros((ni,nv)),    np.zeros((ni,ne)),    np.diag(sk),        np.diag(vk)]])
    
    lhs = - C.vertcat([  Jf_e.T + C.mtimes([Jg_e.T,lk]) + C.mtimes(Jh_e.T,vk),
                    g_e,                              
                    h_e + sk,                          
                    vk*sk - tau                    ])
    
    # Termination condition
    if np.linalg.norm(lhs) < th_1:
        if tau < th_2:
            print 'Solution found!'
            break
        else:
            tau = tau*k_b
        
    sol = np.linalg.solve(KKT,lhs)
    # line-search
    max_ls = 100
    x_step  = sol[:nv]
    l_step  = sol[nv:nv+ne]
    v_step  = sol[nv+ne:nv+ne+ni]
    s_step  = sol[nv+ne+ni:]
    alpha = 1
    k_ls = 0.9
    min_step = 1.0e-8
    for j in range(max_ls):
        v_t = vk + alpha*v_step
        s_t = sk + alpha*s_step
        if (not np.any(v_t <= 0) and not np.any(s_t <= 0)):
            break
        alpha = alpha*k_ls
        if np.linalg.norm(alpha*np.vstack([ v_step,s_step])) < min_step:
            raise Exception('Line search failed! Could not find dual feasible step.')
            
    xk  = xk + alpha*x_step
    lk  = lk + alpha*l_step
    vk  = vk + alpha*v_step
    sk  = sk + alpha*s_step
    
    # Print some results
    print 'Iteration: ',i
    print 'tau = ', tau
    print 'norm(lhs) = ',np.linalg.norm(lhs)
    print 'step size = ', alpha
    iters.append(C.vertcat([xk,lk,vk,sk]))

iters = C.horzcat(iters)
import pylab
pylab.figure()
pylab.subplot(2,1,1)
pylab.plot(iters[:2,:].T)
pylab.grid(True)
pylab.xlabel('iterations')
pylab.ylabel('primal solution')
pylab.legend(('x_1','x_2'))
pylab.subplot(2,1,2)
pylab.plot(iters[2:4,:].T)
pylab.grid(True)
pylab.xlabel('iterations')
pylab.ylabel('dual solution')
pylab.legend(('mu','lambda'))

# Plot feasible set, and iterations
pylab.figure()
xs = np.linspace(-6,6,1000)
pylab.plot(xs,np.sqrt(np.sin(xs)),'r')
pylab.plot(xs,-np.sqrt(np.sin(xs)),'r')
pylab.plot(xs,np.sqrt(4-xs**2),'b')
pylab.plot(xs,-np.sqrt(4-xs**2),'b')

[X,Y] = np.meshgrid(np.linspace(-6,6), np.linspace(-6,6))
Z = (X- 4)**2 + (Y - 4)**2
pylab.contour(X,Y,Z)
pylab.plot(iters[0,:],iters[1,:],'ko')
pylab.title('Iterations in the primal space')
pylab.grid(True)

pylab.show()
